import sys

import click
import deal

deal.activate()

from .constants import MAX_PASSWORD_LENGTH
from .exceptions import DumbValueError
from .pwgen import search


@click.command(
    no_args_is_help=True,
)
@click.version_option()
@click.option(
    "--min-uppercase",
    type=int,
    default=0,
    help="The minimum number of uppercase characters.",
)
@click.option(
    "--min-lowercase",
    type=int,
    default=0,
    help="The minimum number of lowercase characters.",
)
@click.option(
    "--min-digits",
    type=int,
    default=0,
    help="The minimum number of digit characters.",
)
@click.option(
    "--min-specials",
    type=int,
    default=0,
    help="The minimum number of special characters.",
)
@click.option(
    "--blocklist",
    type=str,
    default="""'";""",
    show_default=True,
    help="Characters that may not be in the password.",
)
@click.option(
    "--allow-repeating/--reject-repeating",
    help="Allow or reject repeating characters in the password.",
    default=False,
    show_default=True,
)
@click.argument(
    "length",
    type=click.IntRange(min=1, max=MAX_PASSWORD_LENGTH),
)
def cli(
    length: int,
    min_uppercase: int,
    min_lowercase: int,
    min_digits: int,
    min_specials: int,
    blocklist: str,
    allow_repeating: bool,
) -> int:
    try:
        try_password = search(
            length=length,
            min_uppercase=min_uppercase,
            min_lowercase=min_lowercase,
            min_digits=min_digits,
            min_specials=min_specials,
            blocklist=blocklist,
            allow_repeating=allow_repeating,
        )
    except DumbValueError as ve:
        print(ve)
        return 1

    print(try_password)

    return 0


if __name__ == "__main__":
    sys.exit(cli())



import string
from typing import Any, Iterator


class Candidate:
    def __init__(self, password: str) -> None:
        self._password = password

    def _count_string_type(self, haystack: str) -> int:
        """Return a count of how many characters in the password are part of
        the haystack."""
        return sum(1 for char in self._password if char in haystack)

    @property
    def password(self) -> str:
        return self._password

    @password.setter
    def password(self, new_password: str) -> None:
        self._password = new_password

    @property
    def digits(self) -> int:
        """Return a count of the ASCII digit characters in the password.
        >>> Candidate("").digits
        0
        >>> Candidate("abc").digits
        0
        >>> Candidate("123").digits
        3
        >>> Candidate("0a").digits
        1
        >>> Candidate("0#").digits
        1
        """
        return self._count_string_type(string.digits)

    @property
    def specials(self) -> int:
        """Return a count of the ASCII punctuation characters in the password.
        >>> Candidate("").specials
        0
        >>> Candidate("abc").specials
        0
        >>> Candidate(r"abc%^*.").specials
        4
        >>> Candidate(r"a\\bc").specials
        1
        """
        return self._count_string_type(string.punctuation)

    @property
    def uppers(self) -> int:
        """Return a count of the ASCII uppercase characters in the password.
        >>> Candidate("").uppers
        0
        >>> Candidate("abc").uppers
        0
        >>> Candidate("ABc").uppers
        2
        >>> Candidate("ABC").uppers
        3
        """
        return self._count_string_type(string.ascii_uppercase)

    @property
    def lowers(self) -> int:
        return self._count_string_type(string.ascii_lowercase)

    @property
    def has_duplicates(self) -> bool:
        """Return True if the password has duplicate characters, otherwise
        False.
        >>> Candidate("").has_duplicates
        False
        >>> Candidate("ABC").has_duplicates
        False
        >>> Candidate("ABA").has_duplicates
        True
        >>> Candidate("ABB").has_duplicates
        True
        """
        return (
            any(self._password.count(c) != 1 for c in self._password)
            if len(self._password)
            else False
        )

    @property
    def has_repeating(self) -> bool:
        """Return True if the password has repeating characters, otherwise
        False.
        >>> Candidate("").has_repeating
        False
        >>> Candidate("A").has_repeating
        False
        >>> Candidate("ABA").has_repeating
        False
        >>> Candidate("AAB").has_repeating
        True
        >>> Candidate("ABB").has_repeating
        True
        """
        for index in range(1, len(self._password)):
            if self._password[index] == self._password[index - 1]:
                return True
        else:
            return False

    def copy(self) -> "Candidate":
        """Return a copy of self.
        >>> cd = Candidate("A")
        >>> cp = cd.copy()
        >>> cd == cp
        True
        >>> cp is cd
        False
        """
        return Candidate(self._password)

    def __len__(self) -> int:
        """Return length of the password, for len().
        >>> cd = Candidate("hi")
        >>> len(cd)
        2
        """
        return len(self._password)

    def __str__(self) -> str:
        """Convert to a string, for str().
        >>> cd = Candidate("hi")
        >>> str(cd)
        'hi'
        """
        return self._password

    def __repr__(self) -> str:
        """Convert to a formal string, for repr().

        >>> cd = Candidate("hi")
        >>> repr(cd)
        'dumbpw.candidate.Candidate("hi")'
        """
        return '{}.{}("{}")'.format(
            self.__class__.__module__,
            self.__class__.__qualname__,
            self._password,
        )

    def __iter__(self) -> Iterator[str]:
        return iter(self._password)

    def __getitem__(self, item: int) -> str:
        """Supports subscription of the password.
        >>> cd = Candidate("ABC")
        >>> cd[0]
        'A'
        >>> cd[2]
        'C'
        >>> cd[1:]
        'BC'
        >>> cd[:1]
        'A'
        >>> cd[:-1]
        'AB'
        >>> 'B' in cd
        True
        >>> 'D' not in cd
        True
        """
        return self._password[item]

    def __lt__(self, lvalue: str) -> bool:
        """Supports the '<' comparison operator.
        >>> Candidate("B") < Candidate("A")
        False
        >>> Candidate("A") < Candidate("B")
        True
        >>> Candidate("A") < Candidate("A")
        False
        """
        return bool(self._password < lvalue)

    def __le__(self, lvalue: str) -> bool:
        """Supports the '<=' comparison operator.
        >>> Candidate("B") <= Candidate("A")
        False
        >>> Candidate("A") <= Candidate("B")
        True
        >>> Candidate("A") <= Candidate("A")
        True
        """
        return bool(self._password <= lvalue)

    def __eq__(self, lvalue: Any) -> bool:
        """Supports the '==' comparison operator.
        >>> Candidate("A") == Candidate("B")
        False
        >>> Candidate("A") == "B"
        False
        >>> Candidate("A") == Candidate("A")
        True
        >>> Candidate("A") == "A"
        True
        """
        return bool(str(self) == lvalue)

    def __ne__(self, lvalue: Any) -> bool:
        """Supports the '!=' comparison operator.
        >>> Candidate("A") != Candidate("B")
        True
        >>> Candidate("A") != Candidate("A")
        False
        """
        return bool(self._password != lvalue)

    def __gt__(self, lvalue: str) -> bool:
        """Supports the '>' comparison operator.
        >>> Candidate("B") > Candidate("A")
        True
        >>> Candidate("A") > Candidate("B")
        False
        >>> Candidate("A") > Candidate("A")
        False
        """
        return bool(self._password > lvalue)

    def __ge__(self, lvalue: str) -> bool:
        """Supports the '>=' comparison operator.
        >>> Candidate("B") >= Candidate("A")
        True
        >>> Candidate("A") >= Candidate("B")
        False
        >>> Candidate("A") >= Candidate("A")
        True
        """
        return bool(self._password >= lvalue)

    def __add__(self, new_chars: str) -> "Candidate":
        """Supports the '+' operator.
        >>> Candidate("A") + "B"
        dumbpw.candidate.Candidate("AB")
        >>> Candidate("A") + Candidate("B")
        dumbpw.candidate.Candidate("AB")
        """
        return Candidate(str(self) + str(new_chars))



import secrets

import deal

deal.activate()

from .candidate import Candidate
from .charspace import Charspace
from .constants import MAX_PASSWORD_LENGTH
from .exceptions import DumbValueError


@deal.pre(
    validator=lambda _: _.length <= MAX_PASSWORD_LENGTH,
    exception=DumbValueError,
    message=f"length cannot be greater than {MAX_PASSWORD_LENGTH}.",
)
@deal.pre(
    validator=lambda _: _.min_uppercase
    + _.min_lowercase
    + _.min_digits
    + _.min_specials
    <= _.length,
    exception=DumbValueError,
    message="You cannot request more characters than the password length.",
)
@deal.pre(
    validator=lambda _: _.length <= MAX_PASSWORD_LENGTH,
    exception=DumbValueError,
    message=f"length cannot be greater than {MAX_PASSWORD_LENGTH}.",
)
@deal.pre(
    validator=lambda _: _.length > 0,
    message="length must be greater than zero.",
    exception=DumbValueError,
)
@deal.ensure(
    lambda _: len(_.result) == _.length,
    message="The returned value len must equal the requested length.",
)
def search(
    *,
    length: int,
    min_uppercase: int,
    min_lowercase: int,
    min_digits: int,
    min_specials: int,
    blocklist: str,
    allow_repeating: bool,
) -> str:
    charspace = Charspace(blocklist=blocklist)
    try_password = Candidate("")

    while not all(
        [
            True
            and try_password.uppers >= min_uppercase
            and try_password.lowers >= min_lowercase
            and try_password.digits >= min_digits
            and try_password.specials >= min_specials
            and (allow_repeating or not try_password.has_repeating)
            and len(try_password) == length
        ]
    ):
        try_password = Candidate(
            generate(charset=charspace.charset, length=length)
        )

    return str(try_password)


@deal.has("random")
@deal.pre(
    validator=lambda _: _.length <= MAX_PASSWORD_LENGTH,
    message=f"length cannot be greater than {MAX_PASSWORD_LENGTH}.",
    exception=DumbValueError,
)
@deal.pre(
    validator=lambda _: _.length > 0,
    message="length must be greater than zero.",
    exception=DumbValueError,
)
@deal.pre(
    validator=lambda charset, length: len("".join(charset)) > 0,
    message="charset must have positive len.",
    exception=DumbValueError,
)
@deal.ensure(
    lambda _: len(_.result) == _.length,
    message="The returned value len must equal the requested length.",
)
@deal.ensure(
    lambda charset, length, result: all(
        char in "".join(charset) for char in result
    ),
    message="function return value must be "
    "composed of characters in the charset",
)
def generate(*, charset: str, length: int) -> str:
    """Return a cryptographically secure password of len length using
    characters only from the given charset."""
    return "".join(secrets.choice("".join(charset)) for i in range(length))



import string
from typing import Generator


class Charspace:
    def __init__(
        self,
        blocklist: str = "",
        extras: str = string.punctuation,
        uppers: str = string.ascii_uppercase,
        lowers: str = string.ascii_lowercase,
        digits: str = string.digits,
    ) -> None:
        self._blocklist = blocklist
        self._extras = extras
        self._uppers = uppers
        self._lowers = lowers
        self._digits = digits

    @property
    def base_charset(self) -> str:
        return self._lowers + self._uppers + self._digits + self._extras

    @property
    def charset(self) -> str:
        """De-duplicate the base charset, remove characters that are in the
        blocklist, and return a charset as a string."""
        return "".join(
            {char for char in self.base_charset if char not in self._blocklist}
        )

    @property
    def blocklist(self) -> str:
        return self._blocklist

    @blocklist.setter
    def blocklist(self, new_blocklist: str) -> None:
        self._blocklist = new_blocklist

    @property
    def extras(self) -> str:
        return self._extras

    @extras.setter
    def extras(self, new_extras: str) -> None:
        self._extras = new_extras

    @property
    def lowers(self) -> str:
        return self._lowers

    @lowers.setter
    def lowers(self, new_lowers: str) -> None:
        self._lowers = new_lowers

    @property
    def uppers(self) -> str:
        return self._uppers

    @uppers.setter
    def uppers(self, new_uppers: str) -> None:
        self._uppers = new_uppers

    @property
    def digits(self) -> str:
        return self._digits

    @digits.setter
    def digits(self, new_digits: str) -> None:
        self._digits = new_digits

    def __iter__(self) -> Generator[tuple[str, str], None, None]:
        yield "base_charset", self.base_charset
        yield "charset", self.charset
        yield "blocklist", self.blocklist
        yield "extras", self.extras
        yield "lowers", self.lowers
        yield "uppers", self.uppers
        yield "digits", self.digits



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dumbpw']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.0.3,<9.0.0', 'deal>=4.19.1,<5.0.0']

entry_points = \
{'console_scripts': ['dumbpw = dumbpw.cli:cli']}

setup_kwargs = {
    'name': 'dumbpw',
    'version': '0.3.2',
    'description': 'A bad password generator for bad websites with bad password policies',
    'long_description': 'dumbpw\n======================\n|LANGUAGE| |VERSION| |LICENSE| |MAINTAINED| |CIRCLECI| |MAINTAINABILITY|\n|STYLE|\n\n.. |CIRCLECI| image:: https://img.shields.io/circleci/build/gh/rpdelaney/dumbpw\n   :target: https://circleci.com/gh/rpdelaney/dumbpw/tree/main\n.. |LICENSE| image:: https://img.shields.io/badge/license-Apache%202.0-informational\n   :target: https://www.apache.org/licenses/LICENSE-2.0.txt\n.. |MAINTAINED| image:: https://img.shields.io/maintenance/yes/2022?logoColor=informational\n.. |VERSION| image:: https://img.shields.io/pypi/v/dumbpw\n   :target: https://pypi.org/project/dumbpw\n.. |STYLE| image:: https://img.shields.io/badge/code%20style-black-000000.svg\n   :target: https://github.com/psf/black\n.. |LANGUAGE| image:: https://img.shields.io/pypi/pyversions/dumbpw\n.. |MAINTAINABILITY| image:: https://img.shields.io/codeclimate/maintainability-percentage/rpdelaney/dumbpw\n   :target: https://codeclimate.com/github/rpdelaney/dumbpw\n\nTo create and remember passwords for online services, the best practice for\nmost folks online is to use a password management tool such as `Bitwarden\n<https://bitwarden.com/>`_ to generate long, cryptographically random\npasswords. Then, a very strong passphrase is used to lock the password manager.\n\nUnfortunately, in a misguided attempt to encourage users to choose better\npasswords, many websites and apps have `very bad password policies <https://kottke.org/12/06/the-worlds-worst-password-requirements-list>`_\nthat place restrictions on what sorts of characters must be (or may not be) in\na password. These policies inhibit users from using cryptographically random\npassword generators. In fact, a long, high-entropy password is more likely to\nviolate such rules, which means a security-savvy user may have to attempt\nseveral "random" passwords before one is accepted.\n\nEnter dumbpw. dumbpw allows you to configure a set of rules, and then it will\ngenerate a cryptographically secure password that conforms to those dumb rules.\n\nIf all you need is a password generator, **you should not use this**.\n\nInstallation\n------------\n\n.. code-block :: console\n\n    pip3 install dumbpw\n\nUsage\n-----\n\n.. code-block :: console\n\n    $ dumbpw --help\n    Usage: dumbpw [OPTIONS] LENGTH\n\n    Options:\n      --version                       Show the version and exit.\n      --min-uppercase INTEGER         The minimum number of uppercase characters.\n      --min-lowercase INTEGER         The minimum number of lowercase characters.\n      --min-digits INTEGER            The minimum number of digit characters.\n      --min-specials INTEGER          The minimum number of special characters.\n      --blocklist TEXT                Characters that may not be in the password.\n                                      [default: \'";]\n      --allow-repeating / --reject-repeating\n                                      Allow or reject repeating characters in the\n                                      password.  [default: reject-repeating]\n      --help                          Show this message and exit.\n\nKnown issues\n------------\n* dumbpw uses `secrets <https://docs.python.org/3/library/secrets.html>`_\n  to generate passwords. If the generated string doesn\'t meet the given\n  requirements, dumbpw discards it and generates another, until one passes.\n  Therefore, if you ask dumbpw to generate a long password with high minimums,\n  it will run for a very long time before terminating.\n* Likewise, if your minimums require characters that are banned in the\n  blocklist option, dumbpw will run forever.\n* The author is neither a cryptographer, nor a security expert. There has\n  been no formal, independent, external security review of this software. As\n  explained in the LICENSE, the author assumes no responsibility or liability\n  for your use of this software.\n\n============\nDevelopment\n============\n\nTo install development dependencies, you will need `poetry <https://docs.pipenv.org/en/latest/>`_\nand `pre-commit <https://pre-commit.com/>`_.\n\n.. code-block :: console\n\n    pre-commit install --install-hooks\n    poetry install && poetry shell\n\n`direnv <https://direnv.net/>`_ is optional, but recommended for convenience.\n',
    'author': 'Ryan Delaney',
    'author_email': 'ryan.patrick.delaney@protonmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://pypi.org/project/dumbpw',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
